# anomaly.py

import xarray as xr
import numpy as np
import xeofs
import matplotlib.pyplot as plt
import cartopy
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from cartopy.util import add_cyclic_point#(data, coord=None, axis=-1) 


def compute_weights(data, lat_dim=None):
    if lat_dim==None:
        lat_dim='lat'
    """Compute cosine weights based on latitude."""
    return np.cos(np.deg2rad(data[f'{lat_dim}']))

def calculate_anomaly(data, clim_start=None, climatology_dim='time', clim_end=None, freq='month'):   
    """Calculate anomaly based on climatology."""
    if clim_start is None:
        if freq == 'month':
            return data.groupby(f'{climatology_dim}.{freq}') - data.groupby(f'{climatology_dim}.{freq}').mean((f'{climatology_dim}'))
        elif freq == 'dayofyear':
            return data.groupby(f'{climatology_dim}.{freq}') - data.groupby(f'{climatology_dim}.{freq}').mean((f'{climatology_dim}'))
        else:
            raise ValueError('Frequency must be "month" or "dayofyear".')
    else:
        climatology = data.sel(time=slice(f'{clim_start}-01-01', f'{clim_end}-12-31')).groupby(f'{climatology_dim}.{freq}').mean((f'{climatology_dim}'))
        return data.groupby(f'{climatology_dim}.{freq}') - climatology


# def compute_eofs(data, lat_dim='lat'):

#     ''' This functions computes EOF using eofs module '''

#     coslat = np.cos(np.deg2rad(data.coords[lat_dim].values))
#     weights = np.sqrt(coslat)

#     if weights.ndim == 1 and len(weights) == data.sizes[lat_dim]:
#         # Expand weights for broadcasting
#         weights = weights[..., np.newaxis]  # Add a new axis for broadcasting in EOF calculation
#     else:
#         raise ValueError("Weights dimensions do not align with the data anomaly.")

#     try:
#         if lat_dim not in data.dims:
#             solver = Eof(data)
#         else:
#             solver = Eof(data, weights=weights)
#         return solver
        
#     except Exception as e:
#         print(f"Error creating EOF solver: {e}")
#         return None  # Return None on failure to avoid confusion with the exception message

def line_plot(data, figsize=None, dpi=None, variance_fraction=None, color=None, label=None):

    ''' This function plots the calculated indices for the priliminary check 
        if the user wants.
        data: 1D data to be plot
        figsize: size of the canvas. Tuple e.g. (10, 3) default
        dpi : dots per inches-square. Integer 300 default
        variance_fraction: Variance fraction calculated from the given functions 
        such as in calculate_pdo etc.
        color: color for the line desired, black if None
        label: label for the line to be plotted, e.g. PDO, ENSO etc.'''
    try:
        if data.ndim == 1 and 'time' in data.dims:
            plt.figure(figsize=figsize, dpi=dpi)
            ax = plt.axes()
            if ax==None:
                plt.figure(figsize=(10, 3), dpi=300)
                ax=plt.axes()
            if color==None:
                color='k'   
            data.plot(ax=ax, color=color, lw=1, label=label)
            if variance_fraction is not None:
                y_position = data.max() - (data.max() * 0.05) 
                ax.text(int(str(data.time.data[0])[:4]), y_position, f'{label}_vf: {"%.2f" %variance_fraction}')
            # ax.set_title(f'{label}', loc=title_loc, color=color)
            ax.set_title('', loc='center', color='w')
            ax.legend(frameon=False)
            plt.show()
    except Exception as e:
        return print(f"Error while plotting: {e}")



def contour_plot(data, projection=None, figsize=None, dpi=None, cmap=None, extend=None, levels=None, central_lon=None, central_lat=None, ax_global=False):

    ''' This function plots the calculated variability patterbs for the 
        priliminary check if the user wants.
        data (xr_DataArray): 2D data to be plot
        projection (projection object ccrs.Projection_type()): cartopy projection for contour plot
        cmap (str): colormap
        extend (str): extend colorbar, takes arguments, max, min, both 
        figsize (tuple; e.g. default (12, 6)): size of the canvas.
        dpi (int default 450): dots per inches-square
        central_lat: center of latitude for projection
        central_lon: centre of longitude for projection
        levels (list sorted as min to max.)= contour levels

        '''

    try:
        if data.ndim==2 and 'lat' in data.dims and 'lon' in data.dims:
            if central_lon==None:
                central_lon=0
            if central_lat==None:
                central_lat=0
            if projection==None:
                projection=ccrs.PlateCarree(central_longitude=central_lon)
            plt.figure(figsize=figsize, dpi=dpi)
            ax = plt.axes(projection=projection)
            lon=data.lon
            if levels == None:
                levels=list(np.linspace(data.min(), data.max(), 10))
            if (lon.min() <= -178 and lon.max() >= 178) or (lon.min() <= 2. and lon.max() >= 357.):
                cyclic_data, cyclic_lon = add_cyclic_point(data, coord=lon)

                cyclic_data = xr.DataArray(
                    cyclic_data, 
                    dims=data.dims,  # Keep the same dimensions
                    coords={**data.coords, 'lon': cyclic_lon},  # Replace the original longitude with cyclic longitude
                    attrs=data.attrs  # Copy the original attributes
                )
                if ax_global==True:
                    ax.set_global()
                 
                data = cyclic_data#.plot.contourf(levels=levels, cmap=cmap, extend=extend, transform=ccrs.PlateCarree())
            # else:
            data.plot.contourf(levels=levels, cmap=cmap, extend=extend, transform=ccrs.PlateCarree())
            # ax.coastlines()
            # ax.cfeature()
            ax.add_feature(cfeature.LAND, facecolor='grey')
            ax.add_feature(cfeature.OCEAN, facecolor='w')
            ax.add_feature(cfeature.COASTLINE)
    except Exception as e:
        return print(f"Error while plotting: {e}")



def compute_rotated_eofs(data, rotated=None, n_modes=None, standardize=None, use_coslat=None):
    """
    Compute EOFs using the xeofs module, with optional Varimax or Promax rotation.

    Parameters:
    -----------
    data : xarray.DataArray
        Input data with a time dimension.
    rotated : str, optional
        Specify 'Varimax' for orthogonal rotation or 'Promax' for oblique rotation.
        Leave as None for unrotated EOFs.
    n_modes : int, default=10
        Number of modes to compute.
    standardize : bool, optional
        Whether to standardize the data. Default is False.
    use_coslat : bool, optional
        If True (default), weights EOFs by the cosine of latitude for area-averaging.

    Returns:
    --------
    model : xeofs.single.EOF or xeofs.single.EOFRotator
        The fitted EOF model, either rotated or unrotated.

    Notes:
    ------
    - Varimax corresponds to orthogonal rotations (power=1).
    - Promax corresponds to oblique rotations (power=4).
    - Returns None if an invalid rotation option is provided.
    """
    if rotated==None:
        n_modes = n_modes if n_modes is not None else 1
    else:
        n_modes = n_modes if n_modes is not None else 10
    standardize = standardize if standardize is not None else False
    use_coslat = use_coslat if use_coslat is not None else True
    rotated = rotated if rotated is not None else False

    model = xeofs.single.EOF(n_modes=n_modes, standardize=standardize, use_coslat=use_coslat)
    
    try:
        model.fit(data, dim="time")
        
        # If no rotation is specified, return the fitted EOF model.
        if not rotated:
            return model
        
        # Choose the rotation method based on the input.
        if rotated == 'Varimax':
            rot_model = xeofs.single.EOFRotator(n_modes=n_modes, power=1)  # Varimax: Orthogonal rotation
            return rot_model.fit(model)
        elif rotated == 'Promax':
            rot_model = xeofs.single.EOFRotator(n_modes=n_modes, power=4)  # Promax: Oblique rotation
            return rot_model.fit(model)
        else:
            print("Invalid rotation option. Please specify None, 'Varimax', or 'Promax'.")
            return None
    
    except Exception as e:
        print(f"Error during EOF calculation: {e}")
        return None



import xarray as xr
from xIndices.indices import compute_sst_trend_enso, compute_pdo

# Load SST data
sst = xr.open_dataset('./sst.mnmean.nc')['sst']

# Define climatology period
clim_start = 1981
clim_end = 2010

# Compute SST trend and ENSO
sst_trend_pattern, enso_pattern, sst_trend_ts, enso_index = compute_sst_trend_enso(sst, clim_start, clim_end)

# Compute PDO
pdo_pattern, pdo_index = compute_pdo(sst, clim_start, clim_end)


# __init__.py

from .utils import calculate_anomaly, compute_weights, line_plot, compute_rotated_eofs, contour_plot
from .indices import global_sst_trend_and_enso, compute_pdo, compute_amo, compute_regional_eof_modes
from .preprocess_data import load_data, write_netcdf, regridding, adjust_longitude, rename_dims_to_standard



# setup.py

from setuptools import setup, find_packages
import pathlib

# Read the README file
this_dir = pathlib.Path(__file__).parent
long_description = (this_dir / "README.md").read_text()



setup(
    name='xIndices',
    version='1.2.1',
    description='A xarray based module for computing SST trends and SST variability modes and other atmospheric variability modes',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Jivesh Dixit',
    author_email='jiveshdixit@cas.iitd.ac.in',
    packages=find_packages(),
    classifiers=[
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Operating System :: OS Independent',
    'Topic :: Communications :: Email'],
    install_requires=[
        'xarray',
        'numpy>=1.26, <2.0',
        'dask',
        'xesmf',
        'matplotlib',
        'cartopy',
        'xeofs >2.2.3'
    ],
    python_requires='>=3.10, <3.13'
)



# eof_calculations.py

import numpy as np
from .utils import calculate_anomaly, compute_weights, compute_rotated_eofs
import xarray as xr
from .preprocess_data import load_data, rename_dims_to_standard, adjust_longitude



def calculate_global_mean_sst(data, lat_name='lat', lon_name='lon'):
    return calculate_anomaly(data).weighted(compute_weights(data)).mean(dim=[lat_name, lon_name])



def global_sst_trend_and_enso(data=None, path=None, var=None, clim_start=None, clim_end=None, desired=None, 
    start_time=None, end_time=None, to_range='0_360', standardize=False, 
    normalize_pattern=True, normalize_index=False):
    """
    Calculate global SST warming trend and ENSO patterns using EOF analysis.

    This function computes the global SST warming trend, ENSO index, and their
    corresponding variance fractions using the first two EOF modes of SST anomalies.
    It first removes the annual cycle climatology and then applies EOF analysis.

    Parameters:
    ----------
    data : xarray.DataArray or xarray.Dataset, optional
        Gridded SST data with dimensions (time, lat, lon). If not provided, 
        'path' and 'var' must be used to load the data.
    path : str, optional
        Directory path to the data. Required if 'data' is not provided.
    var : str, optional
        Name of the variable to load from the dataset.
    clim_start : int, optional
        Climatology start year.
    clim_end : int, optional
        Climatology end year (if both None, the entire period is used for climatology).
    desired : list of str, optional
        Desired outputs, which can include:
        ['sst_trend_pattern', 'sst_trend_timeseries', 'variance_fraction_trend', 
        'enso_pattern', 'enso_index', 'variance_fraction_enso']. 
        Default is to return all.
    start_time : int, optional
        Start year for loading data.
    end_time : int, optional
        End year for loading data.
    to_range : str, optional, default '0_360'
        Longitude range for data adjustment. Options are '0_360' or '-180_180'.
    standardize : bool, optional, default False
        Whether to normalize the data with its standard deviation before EOF calculation.
    normalize_pattern : bool, optional, default True
        Whether to normalize spatial components (patterns) with singular values.
    normalize_index : bool, optional, default False
        Whether to normalize time series (scores) with singular values.

    Returns:
    -------
    List containing the desired outputs: SST trend pattern, trend timeseries,
    variance fraction of the trend, ENSO pattern, ENSO index, and variance fraction of ENSO.
    """


    if desired is None:
        desired = [
            'sst_trend_pattern', 'sst_trend_timeseries', 'variance_fraction_trend',
            'enso_pattern', 'enso_index', 'variance_fraction_enso'
        ]
    

    if path and var:
        data = adjust_longitude(
            rename_dims_to_standard(
                load_data(path, var, start_time=start_time, end_time=end_time)
            ),
            to_range=to_range
        )
    

    data_anom = calculate_anomaly(data, clim_start=clim_start, clim_end=clim_end)


    solver = compute_rotated_eofs(
        data_anom, rotated=False, n_modes=2, standardize=standardize, use_coslat=True
    )
    

    eofs_ = solver.components(normalized=normalize_pattern)
    pcs_ = solver.scores(normalized=normalize_index)
    var_frac_ = solver.explained_variance_ratio()


    result_dict = {
        'sst_trend_pattern': eofs_[0].squeeze(),
        'sst_trend_timeseries': (pcs_[0] / pcs_[0].std()).squeeze(),
        'variance_fraction_trend': var_frac_[0].squeeze(),
        'enso_pattern': eofs_[1].squeeze(),
        'enso_index': (pcs_[1] / pcs_[1].std()).squeeze(),
        'variance_fraction_enso': var_frac_[1].squeeze()
    }

    return_desired = [result_dict[key] for key in desired if key in result_dict]
    return return_desired[0] if len(return_desired) == 1 else return_desired



def compute_regional_eof_modes(data=None, path=None, var=None, clim_start=None, clim_end=None, desired=None, 
    start_time=None, end_time=None, lat_s=90, lat_e=-90, lon_s=0, lon_e=360, 
    to_range='0_360', n_modes=1, remove_trend=False, rotated=False, 
    use_coslat=True, standardize=False, normalize_pattern=True, normalize_index=False):
    """
    Calculate regional EOF (Empirical Orthogonal Functions) modes from gridded SST data.

    This generalized function computes EOFs for a specified regional domain, allowing for 
    trend removal, rotation, and different normalization options.

    Parameters:
    ----------
    data : xarray.DataArray or xarray.Dataset, optional
        Gridded SST data with dimensions (time, lat, lon). Use 'path' and 'var' 
        instead if you want to load the data from a file.
    path : str, optional
        Directory path to the data. Required if 'data' is not provided.
    var : str, optional
        Name of the variable to load from the dataset.
    clim_start : int, optional
        Climatology start year.
    clim_end : int, optional
        Climatology end year (if both None, the entire period is used for climatology).
    desired : list, optional
        Desired outputs, which can be ['regional_patterns', 'regional_timeseries', 
        'variance_fractions_regional']. Default is all three.
    start_time : int, optional
        Start year for loading data.
    end_time : int, optional
        End year for loading data.
    lat_s : float, optional
        Start latitude for the regional selection. Default is 90.
    lat_e : float, optional
        End latitude for the regional selection. Default is -90.
    lon_s : float, optional
        Start longitude for the regional selection. Default is 0.
    lon_e : float, optional
        End longitude for the regional selection. Default is 360.
    to_range : str, optional
        Target longitude range. Default is '0_360'. Use '-180_180' if needed.
    n_modes : int, optional
        Number of EOF modes to compute. Default is 1.
    remove_trend : bool, optional
        Whether to remove the global trend before calculating EOFs. Default is False.
    rotated : bool, optional
        Whether to apply Varimax or Promax rotation to the EOFs. Default is False.
    use_coslat : bool, optional
        Whether to use cosine latitude weighting. Default is True.
    standardize : bool, optional
        Whether to normalize the data with its standard deviation before EOF calculation. Default is False.
    normalize_pattern : bool, optional
        Whether to normalize spatial components (patterns) with singular values. Default is True.
    normalize_index : bool, optional
        Whether to normalize the time series (scores) with singular values. Default is False.

    Returns:
    -------
    List containing the desired outputs: regional patterns, timeseries, and/or variance fractions.
    """

    if desired is None:
        desired = ['regional_patterns', 'regional_timeseries', 'variance_fractions_regional']
    

    if path and var:
        data = adjust_longitude(
            rename_dims_to_standard(
                load_data(
                    path, var, start_time=start_time, end_time=end_time, 
                    lat_s=lat_s, lat_e=lat_e, lon_s=lon_s, lon_e=lon_e
                )
            ),
            to_range=to_range
        )
    

    data_anom = calculate_anomaly(data, clim_start=clim_start, clim_end=clim_end)
    if remove_trend:
        global_mean_sst = calculate_global_mean_sst(data, lat_name='lat', lon_name='lon')
        data_anom -= global_mean_sst


    solver = compute_rotated_eofs(
        data_anom, rotated=rotated, n_modes=n_modes, 
        standardize=standardize, use_coslat=use_coslat
    )


    result_dict = {
        'regional_patterns': solver.components(normalized=normalize_pattern).squeeze(),
        'regional_timeseries': (solver.scores(normalized=normalize_index) / 
                                solver.scores(normalized=normalize_index).std()).squeeze(),
        'variance_fractions_regional': solver.explained_variance_ratio().squeeze()
    }


    return_desired = [result_dict[key] for key in desired if key in result_dict]
    return return_desired[0] if len(return_desired) == 1 else return_desired




def compute_pdo(data=None, path=None, var=None, clim_start=None, clim_end=None, desired=None, 
    start_time=None, end_time=None, to_range='0_360', standardize=False, 
    normalize_pattern=True, normalize_index=False, lat_s=70, lat_e=20, lon_s=110, 
    lon_e=260, remove_trend=False):
    """
    Calculate the PDO (Pacific Decadal Oscillation) index and pattern.

    This function calculates the conventional PDO index, pattern, and variance fraction 
    as the first EOF modes of SST anomaly in the North Pacific (above 20°N) after 
    removing the annual cycle climatology and global mean SST (to account for the warming trend).

    Parameters:
    ----------
    data : xarray.DataArray or xarray.Dataset, optional
        Gridded SST data with dimensions (time, lat, lon). Use 'path' and 'var' 
        instead if you want to load the data from a file.
    path : str, optional
        Directory path to the data. Required if 'data' is not provided.
    var : str, optional
        Name of the variable to load from the dataset.
    clim_start : int, optional
        Climatology start year.
    clim_end : int, optional
        Climatology end year (if both None, whole period will be chosen for climatology).
    desired : list, optional
        Desired outputs, which can be ['pdo_pattern', 'pdo_index', 'variance_fraction_pdo']. 
        Default is all three.
    start_time : int, optional
        Start year for loading data.
    end_time : int, optional
        End year for loading data.
    to_range : str, optional
        Target longitude range. Default is '0_360'. Use '-180_180' if needed.
    standardize : bool, optional
        Whether to normalize the data with its standard deviation.
    normalize_pattern : bool, optional
        Whether to normalize spatial components with singular values. Default is True.
    normalize_index : bool, optional
        Whether to normalize the time series (scores) with singular values. Default is False.
    remove_trend : bool, optional
        Whether to remove the global trend. Default is False, in which case mode 2 is used. 
        If True, mode 1 is used.

    Returns:
    -------
    List containing the desired outputs: PDO pattern, PDO index, and/or variance fraction.
    """

    if desired is None:
        desired = ['pdo_pattern', 'pdo_index', 'variance_fraction_pdo']
    

    n_modes = 1 if remove_trend else 2
    

    if path and var:
        data = adjust_longitude(
            rename_dims_to_standard(
                load_data(
                    path, var, start_time=start_time, end_time=end_time, 
                    lat_s=lat_s, lat_e=lat_e, lon_s=lon_s, lon_e=lon_e
                )
            ),
            to_range=to_range
        )


    data_anomaly = calculate_anomaly(data)
    data_pdo = data_anomaly.sel(lat=slice(lat_s, lat_e), lon=slice(lon_s, lon_e))
    

    if remove_trend:
        data_pdo_anomaly = data_pdo - calculate_global_mean_sst(data, lat_name='lat', lon_name='lon')
    else:
        data_pdo_anomaly = calculate_anomaly(data_pdo, clim_start=clim_start, clim_end=clim_end)
    

    solver = compute_rotated_eofs(
        data_pdo_anomaly, rotated=False, n_modes=n_modes, 
        standardize=standardize, use_coslat=True
    )
    

    result_dict = {
        'pdo_pattern': solver.components()[n_modes - 1].squeeze(),
        'pdo_index': (solver.scores()[n_modes - 1] / solver.scores()[n_modes - 1].std()).squeeze(),
        'variance_fraction_pdo': solver.explained_variance_ratio()[n_modes - 1].squeeze()
    }
    

    return_desired = [result_dict[key] for key in desired if key in result_dict]
    return return_desired[0] if len(return_desired) == 1 else return_desired



def compute_amo(data=None, path=None, var=None, clim_start=None, clim_end=None, desired=None, 
    start_time=None, end_time=None, lat_s=70, lat_e=0, lon_s=280, lon_e=360, 
    to_range='0_360'):
    """
    Calculate the AMO (Atlantic Multidecadal Oscillation) index and pattern.

    This function calculates the conventional AMO index and AMO pattern as the 
    area-averaged SST anomalies and time-averaged SST anomalies in the North Atlantic 
    (0°N-70°N) after removing the annual cycle climatology and global mean SST (to 
    account for the warming trend).

    Parameters:
    ----------
    data : xarray.DataArray or xarray.Dataset, optional
        Gridded SST data with dimensions (time, lat, lon). Use 'path' and 'var' 
        instead if you want to load the data from a file.
    path : str, optional
        Directory path to the data. Required if 'data' is not provided.
    var : str, optional
        Name of the variable to load from the dataset.
    start_time : int, optional
        Start year for loading data.
    end_time : int, optional
        End year for loading data.
    clim_start : int, optional
        Climatology start year.
    clim_end : int, optional
        Climatology end year (if both None, whole period will be chosen for climatology).
    desired : list, optional
        Desired outputs, which can be ['amo_pattern', 'amo_index']. Default is both.
    lat_s : float, optional
        Start latitude for the region. Default is 70 (North).
    lat_e : float, optional
        End latitude for the region. Default is 0 (Equator).
    lon_s : float, optional
        Start longitude for the region. Default is 280 (Western Atlantic).
    lon_e : float, optional
        End longitude for the region. Default is 360.
    to_range : str, optional
        Target longitude range. Default is '0_360'. Use '-180_180' if needed.

    Returns:
    -------
    List containing the desired outputs: AMO pattern and/or AMO index.
    """

    if desired is None:
        desired = ['amo_pattern', 'amo_index']


    if path is not None and var is not None:
        data = adjust_longitude(
            rename_dims_to_standard(
                load_data(path, var, start_time=start_time, end_time=end_time)
            ),
            to_range=to_range
        )


    north_atlantic_sst = data.sel(lat=slice(lat_s, lat_e), lon=slice(lon_s, lon_e))
    

    data_anomalies = calculate_anomaly(
        north_atlantic_sst, clim_start=clim_start, clim_end=clim_end
    )
    

    global_mean_data = calculate_global_mean_sst(data, lat_name='lat', lon_name='lon')
    

    amo_index = (
        data_anomalies.weighted(compute_weights(data_anomalies)).mean(('lat', 'lon')) 
        - global_mean_data
    )
    

    amo_pattern = (
        xr.cov(calculate_anomaly(data), amo_index, dim='time') 
        / amo_index.var()
    )
    

    result_dict = {
        'amo_pattern': amo_pattern,
        'amo_index': amo_index
    }
    

    return_desired = [result_dict[key] for key in desired if key in result_dict]
    return return_desired[0] if len(return_desired) == 1 else return_desired



def compute_nao(data=None, path=None, var=None, clim_start=None, clim_end=None, desired=None, 
    lat_s=None, lat_e=None, use_coslat=None, standardize=None, to_range=None, n_modes=10, nao_mode=None,
    start_time=None, end_time=None, rotated='Varimax'):
    '''
    This function calculates the NAO index, NAO pattern, and variance fraction.
    It is calculated as the second EOF mode of 500mb geopotential height 
    anomaly above 20N after removing the annual cycle climatology.

    Parameters:
    - data : xarray.DataArray or xarray.Dataset
        Processed gridded data in the shape of (time, lat, lon).
    
    - path : str, optional
        Path to the data file (if data is to be loaded directly).
    
    - var : str, optional
        Variable name in the dataset (used if loading from a file).
    
    - clim_start : int, optional
        Climatology start year.
    
    - clim_end : int, optional
        Climatology end year. If None, the whole period will be used.
    
    - desired : list, optional
        List of desired outputs, e.g., ['nao_pattern', 'nao_index', 'variance_fraction_nao'].
    
    - lat_s : float, optional
        Start latitude for selecting the northern region.
    
    - lat_e : float, optional
        End latitude for selecting the northern region.
    
    - use_coslat : bool, optional
        Whether to apply cosine latitude weighting. Default is True.
    
    - standardize : bool, optional
        If True, standardizes the data. Default is False.
    
    - to_range : str, optional
        Longitude range, either '0_360' or '-180_180'. Default is '0_360'.
    
    - nao_mode : int, optional
        The EOF mode number to be considered as NAO. Default is 1.
    
    - start_time, end_time : str, optional
        Time range for selecting data.
    
    - rotated : str, optional
        Rotation method for EOFs. Options: None, 'Varimax', 'Promax'. Default is 'Varimax'.
    '''
    

    lat_s = lat_s if lat_s is not None else 90
    lat_e = lat_e if lat_e is not None else 20
    standardize = standardize if standardize is not None else False
    desired = desired if desired is not None else ['nao_pattern', 'nao_index', 'variance_fraction_nao']
    to_range = to_range if to_range is not None else '0_360'
    nao_mode = nao_mode if nao_mode is not None else 1


    if path is not None and var is not None:
        data = adjust_longitude(rename_dims_to_standard(
            load_data(path, var, start_time=start_time, end_time=end_time, lat_s=lat_s, lat_e=lat_e)
        ), to_range=to_range)


    if data is None:
        raise ValueError("Data must be provided either directly or via path and var.")

    north_geop = data.sel(lat=slice(lat_s, lat_e))
    

    data_anomalies = calculate_anomaly(north_geop, clim_start=clim_start, clim_end=clim_end)
    

    eofs_result = compute_rotated_eofs(
        data_anomalies, rotated=rotated, n_modes=n_modes, 
        standardize=standardize, use_coslat=use_coslat
    )


    nao_index = eofs_result.scores()[nao_mode-1] / eofs_result.scores()[nao_mode-1].std()
    nao_pattern = eofs_result.components()[nao_mode-1]
    variance_fraction_nao = eofs_result.explained_variance_ratio()[nao_mode-1]
    
    result_dict = {
        'nao_pattern': nao_pattern.squeeze(),
        'nao_index': nao_index.squeeze(),
        'variance_fraction_nao': variance_fraction_nao.squeeze(),
    }
    

    return_desired = [result_dict[key] for key in desired if key in result_dict]
    return return_desired[0] if len(return_desired) == 1 else return_desired

